@@ -84,6 +84,11 @@ static void handle_panel_keybinding   (MetaDisplay    *display,
                                        MetaWindow     *window,
                                        XEvent         *event,
                                        MetaKeyBinding *binding);
+static void handle_toggle_above       (MetaDisplay    *display,
+                                       MetaScreen     *screen,
+                                       MetaWindow     *window,
+                                       XEvent         *event,
+                                       MetaKeyBinding *binding);
 static void handle_toggle_maximize    (MetaDisplay    *display,
                                        MetaScreen     *screen,
                                        MetaWindow     *window,
@@ -353,6 +358,7 @@ static const MetaKeyHandler screen_handlers[] = {
 static const MetaKeyHandler window_handlers[] = {
   { META_KEYBINDING_WINDOW_MENU, handle_activate_menu, NULL },
   { META_KEYBINDING_TOGGLE_FULLSCREEN, handle_toggle_fullscreen, NULL },
+  { META_KEYBINDING_TOGGLE_ABOVE, handle_toggle_above, NULL },
   { META_KEYBINDING_TOGGLE_MAXIMIZE, handle_toggle_maximize, NULL },
   { META_KEYBINDING_MAXIMIZE, handle_maximize, NULL },
   { META_KEYBINDING_UNMAXIMIZE, handle_unmaximize, NULL },
@@ -925,7 +931,7 @@ meta_change_keygrab (MetaDisplay *display,
                      unsigned int keycode,
                      int          modmask)
 {
-  int ignored_mask;
+  unsigned int ignored_mask;
 
   /* Grab keycode/modmask, together with
    * all combinations of ignored modifiers.
@@ -942,7 +948,7 @@ meta_change_keygrab (MetaDisplay *display,
   meta_error_trap_push (display);
   
   ignored_mask = 0;
-  while (ignored_mask < (int) display->ignored_modifier_mask)
+  while (ignored_mask <= display->ignored_modifier_mask)
     {
       if (ignored_mask & ~(display->ignored_modifier_mask))
         {
@@ -2195,10 +2201,6 @@ process_tab_grab (MetaDisplay *display,
       
       if (target_window)
         {
-          meta_topic (META_DEBUG_KEYBINDINGS,
-                      "Ending grab early so we can focus the target window\n");
-          meta_display_end_grab_op (display, event->xkey.time);
-
           meta_topic (META_DEBUG_KEYBINDINGS,
                       "Activating target window\n");
 
@@ -2206,6 +2208,10 @@ process_tab_grab (MetaDisplay *display,
                       target_window->desc);
           meta_window_activate (target_window, event->xkey.time);
 
+          meta_topic (META_DEBUG_KEYBINDINGS,
+                      "Ending grab early so we can focus the target window\n");
+          meta_display_end_grab_op (display, event->xkey.time);
+
           return TRUE; /* we already ended the grab */
         }
       
@@ -2603,7 +2609,7 @@ handle_toggle_desktop (MetaDisplay    *display,
   if (screen->showing_desktop)
     {
       meta_screen_unshow_desktop (screen);
-      meta_screen_focus_top_window (screen, NULL);
+      meta_workspace_focus_top_window (screen->active_workspace, NULL);
     }
   else
     meta_screen_show_desktop (screen);
@@ -2860,6 +2866,22 @@ handle_toggle_fullscreen  (MetaDisplay    *display,
     }
 }
 
+static void
+handle_toggle_above       (MetaDisplay    *display,
+                           MetaScreen     *screen,
+                           MetaWindow     *window,
+                           XEvent         *event,
+                           MetaKeyBinding *binding)
+{
+  if (window)
+    {
+      if (window->wm_state_above)
+        meta_window_unmake_above (window);
+      else
+        meta_window_make_above (window);
+    }
+}
+
 static void
 handle_toggle_maximize    (MetaDisplay    *display,
                            MetaScreen     *screen,
@@ -3022,7 +3044,7 @@ do_handle_move_to_workspace  (MetaDisplay    *display,
       /* Activate second, so the window is never unmapped */
       meta_window_change_workspace (window, workspace);
       if (flip)
-        meta_workspace_activate (workspace);
+        meta_workspace_activate_with_focus (workspace, window);
     }
   else
     {
